# -*- coding: UTF-8 -*-

import dataclasses
import typing
from collections import UserString

import orjson
from starlette.responses import Response, StreamingResponse

type BaseTypeAnnotation = str | int | float | bool | None


@dataclasses.dataclass(slots=True, kw_only=True)
class _CustomDataType:
    type_class: type
    encoder: typing.Callable[[typing.Any], BaseTypeAnnotation]


_CUSTOM_DATA_TYPE_LIST = [_CustomDataType(type_class=UserString, encoder=str)]


def register_custom_data_types(
        *, type_class: type,
        encoder: typing.Callable[[typing.Any], BaseTypeAnnotation],
) -> None:
    for custom_data_type in _CUSTOM_DATA_TYPE_LIST:
        if custom_data_type.type_class == type_class:
            raise ValueError(f'Type class {type_class} already registered')
    _CUSTOM_DATA_TYPE_LIST.append(
        _CustomDataType(type_class=type_class, encoder=encoder))


def orjson_dumps_default(obj):
    for custom_data_type in _CUSTOM_DATA_TYPE_LIST:
        if isinstance(obj, custom_data_type.type_class):
            return custom_data_type.encoder(obj)
    raise TypeError(f'Object of type {type(obj)} is not JSON serializable')


class ORJSONResponse(Response):
    media_type = 'application/json'

    def render(self, content: typing.Any) -> bytes:
        return orjson.dumps(
            content, option=orjson.OPT_STRICT_INTEGER,
            default=orjson_dumps_default)


class StreamResponse(StreamingResponse):
    pass



# -*- coding: UTF-8 -*-

import orjson
from marshmallow import Schema
from starlette.requests import Request as StarletteRequest


class MusoRequest:

    def __init__(
            self,
            starlette_request: StarletteRequest,
            query_args_schema: Schema | None = None,
            form_data_schema: Schema | None = None,
            json_body_schema: Schema | None = None,
    ):
        self._starlette_request = starlette_request

        self._query_args_schema: Schema | None = query_args_schema
        self._form_data_schema: Schema | None = form_data_schema
        self._json_body_schema: Schema | None = json_body_schema

        self._cached_query_args: dict | None = None
        self._cached_form_data: dict | None = None
        self._cached_json_body: dict | None = None

    async def query_args(self) -> dict:
        if self._cached_query_args is None:
            self._cached_query_args = self._query_args_schema.load(
                self._starlette_request.query_params) or dict()
        return self._cached_query_args

    async def form_data(self) -> dict:
        if self._cached_form_data is None:
            form_data = await self._starlette_request.form()
            self._cached_form_data = self._form_data_schema.load(
                form_data) or dict()
        return self._cached_form_data

    async def json_body(self) -> dict:
        if not hasattr(self._starlette_request, '_json'):
            body = await self._starlette_request.body()
            json_result = orjson.loads(body)
            self._starlette_request._json = json_result    # noqa
            self._cached_json_body = self._json_body_schema.load(json_result)
        return self._cached_json_body

    def __getattr__(self, item):
        return getattr(self._starlette_request, item)



# -*- coding: UTF-8 -*-

import abc
import dataclasses
import datetime
import functools
import inspect
import typing

import orjson
from psycopg import AsyncConnection, AsyncCursor
from psycopg.rows import Row
from psycopg.sql import SQL
from psycopg.types.json import set_json_dumps, set_json_loads
from psycopg_pool import AsyncConnectionPool

set_json_loads(orjson.loads)
set_json_dumps(orjson.dumps)

_NAMED_POOL_DICT: dict[str, AsyncConnectionPool] = dict()

type ConfigureFunc = typing.Callable[
    [AsyncConnection], typing.Awaitable | None]


async def init(
        *, name: str, dsn: str, pool_size: int = 16, open_timeout: int = 15,
        configure_funcs: typing.Sequence[ConfigureFunc] = None,
) -> None:
    if name in _NAMED_POOL_DICT:
        raise ValueError(f'Pool "{name}" already exists')
    if pool_size < 1:
        raise ValueError('Pool size should be greater than 0')
    minsize, maxsize = pool_size, pool_size
    if minsize > 4:
        minsize = 4
    kwargs = dict(autocommit=False)
    init_kwargs = dict()
    if configure_funcs:
        async def _configure(conn: AsyncConnection):
            for func in configure_funcs:
                if inspect.iscoroutinefunction(func):
                    await func(conn)
                else:
                    func(conn)

        init_kwargs['configure'] = _configure
    pool = AsyncConnectionPool(
        conninfo=dsn, min_size=minsize, max_size=maxsize, open=False,
        name=name, kwargs=kwargs, **init_kwargs)
    await pool.open(wait=True, timeout=open_timeout)
    _NAMED_POOL_DICT[name] = pool


async def close_all():
    for _, pool in _NAMED_POOL_DICT.items():
        await pool.close()
    _NAMED_POOL_DICT.clear()


def with_postgres(*, name: str, transaction: bool = False):
    def wrapper(func):
        argspec = inspect.getfullargspec(func)
        if all(map(lambda x: 'cursor' not in x,
                   (argspec.args, argspec.kwonlyargs))):
            raise SyntaxError('`cursor` is a required argument')

        @functools.wraps(func)
        async def wrapped(*args, **kwargs):
            if name not in _NAMED_POOL_DICT:
                raise SyntaxError(f'Pool "{name}" not found')
            if 'cursor' in kwargs:
                raise SyntaxError('`cursor` is a reserved argument')
            async with _NAMED_POOL_DICT[name].connection() as conn:
                async with conn.cursor() as cursor:
                    kwargs['cursor'] = cursor
                    if transaction:
                        async with conn.transaction():
                            result = await func(*args, **kwargs)
                    else:
                        result = await func(*args, **kwargs)
            return result

        return wrapped

    return wrapper


@dataclasses.dataclass
class PostgresBaseModel:
    id_: int
    removed: bool
    created_at: datetime.datetime
    updated_at: datetime.datetime


def make_sql(*, sql_str: str) -> SQL:
    return SQL(sql_str.strip())  # noqa


def _gen_fields(dataclass: type) -> tuple[list[str], list[str]]:
    default_fields = list()
    if issubclass(dataclass, PostgresBaseModel):
        default_fields.extend(
            i.strip('_')
            for i in inspect.get_annotations(PostgresBaseModel).keys())
    custom_fields = list(
        i.strip('_') for i in inspect.get_annotations(dataclass).keys())
    return default_fields, custom_fields


def gen_select_base(*, dataclass: type, table_name: str) -> str:
    default_fields, custom_fields = _gen_fields(dataclass)
    return (f'SELECT {", ".join(default_fields + custom_fields)} '
            f'FROM {table_name}')


def gen_get(*, dataclass: type, table_name: str) -> SQL:
    default_fields, custom_fields = _gen_fields(dataclass)
    return make_sql(sql_str=(
        f'SELECT {", ".join(default_fields + custom_fields)} '
        f'FROM {table_name} WHERE id=%s'))


def gen_insert(*, dataclass: type, table_name: str) -> SQL:
    _, custom_fields = _gen_fields(dataclass)
    return make_sql(sql_str=(
        f'INSERT INTO {table_name} ({", ".join(custom_fields)}) '
        f'VALUES ({", ".join(["%s"] * len(custom_fields))})'))


def gen_insert_returning(*, dataclass: type, table_name: str) -> SQL:
    default_fields, custom_fields = _gen_fields(dataclass)
    return make_sql(sql_str=(
        f'INSERT INTO {table_name} ({", ".join(custom_fields)}) '
        f'VALUES ({", ".join(["%s"] * len(custom_fields))}) '
        f'RETURNING {", ".join(default_fields + custom_fields)}'))


def gen_remove(*, table_name: str) -> SQL:
    return make_sql(sql_str=(
        f'UPDATE {table_name} SET removed=%s, updated_at=NOW() '
        f'WHERE id=%s AND removed=%s'))


@dataclasses.dataclass(slots=True, kw_only=True)
class BaseInsertModel(metaclass=abc.ABCMeta):

    @property
    @abc.abstractmethod
    def as_sql_params(self) -> tuple:
        ...


type DefaultCursorAnnotation = AsyncCursor[AsyncConnection['Row'], 'Row']


class BaseDataFetcher[_T](metaclass=abc.ABCMeta):
    _MODEL_CLASS: type[_T]

    _TABLE_NAME: str
    _SELECT_BASE: str

    _sql_get: SQL

    @classmethod
    async def get(
            cls, *, cursor: DefaultCursorAnnotation, id_: int,
    ) -> _T | None:
        await cursor.execute(query=cls._sql_get, params=(id_,))
        result: Row | None = await cursor.fetchone()
        return cls._MODEL_CLASS(*result) if result else None

    _sql_remove: SQL

    @classmethod
    async def remove(
            cls, *, cursor: DefaultCursorAnnotation, id_: int,
    ) -> None:
        await cursor.execute(query=cls._sql_remove, params=(True, id_, False))



# -*- coding: UTF-8 -*-

import dataclasses
from collections import UserString


@dataclasses.dataclass(slots=True)
class ErrorMeta:
    code: int
    msg: str | UserString
    data: dict | None
    http_status: int
    alert_sentry: bool


class BizError(Exception):
    """ 业务异常基类 """

    __slots__ = ('code', 'msg', 'data', 'http_status', 'alert_sentry', 'args')

    _error = ErrorMeta(1000, '内部错误', None, 500, True)

    def __init__(self, code: int = None, msg: str = None, data: dict = None,
                 http_status: int = None, alert_sentry: bool = None):
        super(BizError, self).__init__()
        self.code = code or self._error.code
        self.msg = msg or self._error.msg
        self.data = data or self._error.data
        self.http_status = http_status or self._error.http_status
        self.alert_sentry = (alert_sentry
                             if alert_sentry is not None
                             else self._error.alert_sentry)
        args = dict(
            code=self.code,
            msg=self.msg,
            data=self.data,
            http_status=self.http_status,
            alert_sentry=self.alert_sentry,
        )
        self.args = (
            '<{e}>'.format(
                e=', '.join(
                    ['{}: {}'.format(k, v) for k, v in args.items()]),
            ),
        )

    def __repr__(self) -> str:
        return '<{class_name}({props})>'.format(
            class_name=self.__class__.__name__,
            props=','.join(
                ['{}={}'.format(k, f'"{v}"' if isinstance(v, str) else v)
                 for k, v in self.__dict__.items()]))
